// Simple Sticky Header and Column implementation for HTML tables
// (c) Vitaliy Filippov 2016
// License: MPL 2.0+
// Version: 2016-12-06

// USAGE:
// makeStickyHeaders(table): add sticky header and footer to table
//   table.parentNode is assumed to be the scroll container
// checkStickyHeaders(table): call this after making table visible
//   (if you've hidden it before)
// initStickyRows(table, trs): initialise table rows (trs = array[TableRow]) if not yet
// fixStickyRows(table, trs): fix table row (trs = array[TableRow]) after modifications
// fixStickyHeader(table): fix table header after modifications
// fixStickyColumnSizes(table, force): adjust column sizes if something is modified
//   if force == true, reposition all fixed cells (header row and column)

// WARNING 1: after using this plugin all columns and rows except 0th ones become shifted by 1.
// table.rows[1] and table.rows[*].cells[1] become special invisible rows/cells with copies of first
// row and column content inside.

// WARNING 2: this plugin does NOT support border-collapse because it is implemented by non-integer border
// sizes O_o and slightly differently in different browsers which leads to problems calculating cell positions.

// WARNING 3: for these hacks to work correctly, you need 'position: relative' on your table and table rows.

// WARNING 4: all table cells should obviously have non-transparent background
// because fixed ones are positioned over normal.

// WARNING 5: make sure you have no CSS rules that will differ on the original header row/column and cloned one,
// and that min/max width/height rules apply to the second row/column, not the first header one.

;(function() {

var _scri = 0;

function getOffsetRect(e)
{
    var b = e.getBoundingClientRect();
    return {
        left: Math.round(b.left*100)/100,
        top: Math.round(b.top*100)/100,
        width: Math.round((b.right-b.left)*100)/100,
        height: Math.round((b.bottom-b.top)*100)/100
    };
}

window.makeStickyHeaders = function(table)
{
    if (!table._scrstyle)
    {
        table._scrstyle = document.createElement('style');
        (document.head||document.getElementsByTagName('head')[0]).appendChild(table._scrstyle);
        table._scri = _scri++;
    }
    if (!table.rows.length || !table.rows[0].children.length)
        return;
    var w = [], l = [], h = table.rows[0].offsetHeight;
    var sr = table.rows[0].cloneNode();
    sr.appendChild(document.createElement('td'));
    sr.children[0].style.width = '0';
    sr.children[0].style.display = 'block';
    sr.children[0].style.position = 'absolute';
    sr.children[0].style.padding = '0';
    sr.style.visibility = 'hidden';
    table._widths = [];
    table._sizerow = sr;
    table.rows[0].style.height = '0';
    table.rows[0].style.display = 'block';
    table.rows[0].style.position = 'absolute';
    table.rows[0].parentNode.insertBefore(sr, table.rows[0].nextSibling);
    fixStickyHeader(table, true);
    table._fixPending = true;
    checkStickyHeaders(table);
    var row0top = getOffsetRect(table.rows[0]).top-getOffsetRect(table.rows[0].offsetParent).top;
    addListener(table.parentNode, 'scroll', function(e)
    {
        var l = this.scrollLeft, t = this.scrollTop;
        table.rows[0].style.top = (t+row0top)+'px';
        if (!table._scrlastleft || table._scrlastleft != l)
        {
            table._scrstyle.innerHTML = l > 0 ? '._scri'+table._scri+' { left: '+l+'px; }' : '';
            table._scrlastleft = l;
        }
    });
}

// should be called after showing table to recalculate fixed cell sizes
window.checkStickyHeaders = function(table)
{
    if (table._fixPending)
    {
        delete table._fixPending;
        var rows = [];
        for (var i = 2; i < table.rows.length; i++)
            rows.push(table.rows[i]);
        fixStickyRows(table, rows);
    }
}

// check if row(s) are not yet initialised and fix them
window.initStickyRows = function(table, trs, nocols)
{
    var ntrs = [];
    for (var i = 0; i < trs.length; i++)
        if (trs[i].children[0].className.indexOf(' _scri'+table._scri) < 0)
            ntrs.push(trs[i]);
    if (ntrs.length)
        fixStickyRows(table, ntrs, nocols);
}

// handle non-header row(s) change with O(1) reflow count
window.fixStickyRows = function(table, trs, nocols)
{
    // (1) make some changes
    var e, d;
    for (var i = 0; i < trs.length; i++)
    {
        e = trs[i].children[0];
        if (e.className.indexOf(' _scri'+table._scri) >= 0)
        {
            // refix
            trs[i].removeChild(trs[i].children[1]);
            d = e.cloneNode(true);
            d.className = d.className.replace(' _scri'+table._scri, '');
            d.style.visibility = 'hidden';
            d.style.width = e._oldWidth;
            d.style.height = '';
            d.style.position = '';
            d.style.display = '';
            d.style.zIndex = '';
            e.parentNode.insertBefore(d, e.nextSibling);
        }
        else
        {
            // fix new
            d = e.cloneNode(true);
            d.style.visibility = 'hidden';
            e.style.position = 'absolute';
            if (e.rowSpan > 1)
                e.rowSpan = 1;
            if (e.style.display == 'none')
                e.style.visibility = 'hidden';
            e.style.display = 'block';
            e.style.zIndex = '1';
            trs[i].insertBefore(d, e.nextSibling);
            e.className += ' _scri'+table._scri;
        }
    }
    if (table.offsetParent)
    {
        // (2) <REFLOW> caused by offsetWidth/offsetHeight
        // (3) remember sizes
        var w = [], h = [], b;
        for (var i = 0; i < trs.length; i++)
        {
            e = trs[i].children[1];
            b = getOffsetRect(e);
            w[i] = b.width;
            h[i] = b.height;
        }
        // (4) apply sizes
        for (var i = 0; i < trs.length; i++)
        {
            e = trs[i].children[0];
            e.style.width = w[i]+'px';
            e.style.height = h[i]+'px';
        }
    }
    else
    {
        // table is invisible
        table._fixPending = true;
    }
    if (!nocols)
        fixStickyColumnSizes(table);
}

window.fixStickyColumnSizes = function(table, force)
{
    if (!table.offsetParent)
    {
        // table is invisible
        table._fixPending = true;
        return;
    }
    // check&fix column sizes
    var changed = false, resizeFixed = false;
    for (var i = 1; i < table._sizerow.children.length; i++)
    {
        var ne = table._sizerow.children[i];
        var nw = getOffsetRect(ne).width;
        var cw = table._widths[i-1];
        if (nw != cw || force)
        {
            table._widths[i-1] = nw;
            changed = true;
            if (i == 1)
                resizeFixed = true;
        }
    }
    if (changed)
    {
        // reposition fixed header
        var w = [], h = [], l = [], b, pb = getOffsetRect(table._sizerow);
        for (var i = 0; i < table._sizerow.children.length; i++)
        {
            b = getOffsetRect(table._sizerow.children[i]);
            w[i] = b.width;
            h[i] = b.height;
            l[i] = b.left-pb.left;
        }
        b = table.rows[0].children[0];
        b.style.width = w[1]+'px';
        b.style.height = h[1]+'px';
        for (var i = 1; i < table._sizerow.children.length; i++)
        {
            b = table.rows[0].children[i];
            b.style.left = l[i]+'px';
            b.style.width = w[i]+'px';
            b.style.height = h[i]+'px';
        }
        if (resizeFixed)
        {
            // resize fixed column
            var h = [];
            for (var i = 0; i < table.rows.length; i++)
                h[i] = getOffsetRect(table.rows[i].children[1]).height;
            for (var i = 0; i < table.rows.length; i++)
            {
                if (i != 1)
                {
                    table.rows[i].children[0].style.width = w[1]+'px';
                    table.rows[i].children[0].style.height = h[i == 0 ? 1 : i]+'px';
                }
            }
        }
    }
}

// handle header row change
window.fixStickyHeader = function(table, nocols)
{
    var tr = table.rows[0];
    var sr = table._sizerow;
    while (sr.children[1])
        sr.removeChild(sr.children[1]);
    var b, e, renew = false;
    for (var i = 0; i < tr.children.length; i++)
    {
        b = tr.children[i];
        e = b.cloneNode(true);
        if (b._sticky)
        {
            if (i == 0)
                e.className = e.className.replace(' _scri'+table._scri, '');
            e.style.position = '';
            e.style.width = b._oldWidth;
            e.style.height = '';
            e.style.left = '';
            e.style.display = '';
            e.style.zIndex = '';
            if (i == 0)
                i++;
        }
        else
            renew = true;
        sr.appendChild(e);
    }
    if (renew)
    {
        for (var i = 0; i < tr.children.length; i++)
        {
            var e = tr.children[i];
            if (!e._sticky)
            {
                e.style.position = 'absolute';
                e.style.display = 'block';
                e.style.zIndex = '1';
                if (i == 0)
                {
                    var d = e.cloneNode(true);
                    d.style.visibility = 'hidden';
                    tr.insertBefore(d, e.nextSibling);
                    i++;
                    e.className += ' _scri'+table._scri;
                    e.style.zIndex = '2';
                }
            }
        }
        table._widths = [];
    }
    if (!nocols)
        fixStickyColumnSizes(table);
}

})();
